"""Simplified command-line interface for Git Smart Squash."""

import argparse
import sys
import subprocess
import json
import os
from typing import List, Dict, Any, Optional
from rich.console import Console
from rich.panel import Panel
from rich.progress import Progress, SpinnerColumn, TextColumn

from .simple_config import ConfigManager
from .ai.providers.simple_unified import UnifiedAIProvider
from .diff_parser import parse_diff, Hunk
from .hunk_applicator import apply_hunks_by_ids, reset_to_base_branch, unstage_all_hunks, HunkApplicatorError


class GitSmartSquashCLI:
    """Simplified CLI for git smart squash."""
    
    def __init__(self):
        self.console = Console()
        self.config_manager = ConfigManager()
        self.config = None
    
    def main(self):
        """Main entry point for the CLI."""
        parser = self.create_parser()
        args = parser.parse_args()
        
        try:
            # Load configuration
            self.config = self.config_manager.load_config(args.config)
            
            # Override config with command line arguments
            if args.ai_provider:
                self.config.ai.provider = args.ai_provider
                # If provider is changed but no model specified, use provider default
                if not args.model:
                    self.config.ai.model = self.config_manager._get_default_model(args.ai_provider)
            if args.model:
                self.config.ai.model = args.model
            
            # Run the simplified smart squash
            self.run_smart_squash(args)
                
        except Exception as e:
            self.console.print(f"[red]Error: {e}[/red]")
            sys.exit(1)
    
    def create_parser(self) -> argparse.ArgumentParser:
        """Create the simplified argument parser."""
        parser = argparse.ArgumentParser(
            prog='git-smart-squash',
            description='AI-powered git commit reorganization for clean PR reviews'
        )
        
        parser.add_argument(
            '--base',
            default='main',
            help='Base branch to compare against (default: main)'
        )
        
        parser.add_argument(
            '--dry-run',
            action='store_true',
            help='Show proposed commit structure without applying'
        )
        
        parser.add_argument(
            '--ai-provider',
            choices=['openai', 'anthropic', 'local', 'gemini'],
            help='AI provider to use'
        )
        
        parser.add_argument(
            '--model',
            help='AI model to use'
        )
        
        parser.add_argument(
            '--config',
            help='Path to configuration file'
        )
        
        return parser
    
    def run_smart_squash(self, args):
        """Run the simplified smart squash operation."""
        try:
            # Ensure config is loaded
            if self.config is None:
                self.config = self.config_manager.load_config()
            
            # 1. Get the full diff between base branch and current branch
            full_diff = self.get_full_diff(args.base)
            if not full_diff:
                self.console.print("[yellow]No changes found to reorganize[/yellow]")
                return
            
            # 2. Parse diff into individual hunks
            hunks = parse_diff(full_diff)
            if not hunks:
                self.console.print("[yellow]No parseable hunks found in diff[/yellow]")
                return
            
            # Create hunk mapping for easy lookup
            hunk_map = {hunk.id: hunk for hunk in hunks}
            
            # 3. Send hunks to AI for commit organization
            with Progress(
                SpinnerColumn(),
                TextColumn("[progress.description]{task.description}"),
                console=self.console
            ) as progress:
                task = progress.add_task("Analyzing changes with AI...", total=None)
                commit_plan = self.analyze_with_ai_hunks(hunks)
            
            if not commit_plan:
                self.console.print("[red]Failed to generate commit plan[/red]")
                return
            
            # 4. Display the plan
            self.display_commit_plan_hunks(commit_plan, hunk_map)
            
            # 5. Execute or dry run
            if args.dry_run:
                self.console.print("\n[green]Dry run complete. Use without --dry-run to apply changes.[/green]")
            else:
                if self.get_user_confirmation():
                    self.apply_commit_plan_hunks(commit_plan, hunk_map, args.base)
                else:
                    self.console.print("Operation cancelled.")
                    
        except Exception as e:
            self.console.print(f"[red]Error: {e}[/red]")
            sys.exit(1)
    
    def get_full_diff(self, base_branch: str) -> Optional[str]:
        """Get the full diff between base branch and current branch."""
        try:
            # First check if we're in a git repo and the base branch exists
            subprocess.run(['git', 'rev-parse', '--git-dir'], 
                         check=True, capture_output=True)
            
            # Try to get the diff
            result = subprocess.run(
                ['git', 'diff', f'{base_branch}...HEAD'],
                capture_output=True, text=True, check=True
            )
            
            if not result.stdout.strip():
                return None
                
            return result.stdout
            
        except subprocess.CalledProcessError as e:
            if 'unknown revision' in e.stderr:
                # Try with origin/main or other common base branches
                for alt_base in [f'origin/{base_branch}', 'develop', 'origin/develop']:
                    try:
                        result = subprocess.run(
                            ['git', 'diff', f'{alt_base}...HEAD'],
                            capture_output=True, text=True, check=True
                        )
                        if result.stdout.strip():
                            self.console.print(f"[yellow]Using {alt_base} as base branch[/yellow]")
                            return result.stdout
                    except subprocess.CalledProcessError:
                        continue
            raise Exception(f"Could not get diff from {base_branch}: {e.stderr}")
    
    def analyze_with_ai_hunks(self, hunks: List[Hunk]) -> Optional[List[Dict[str, Any]]]:
        """Send hunks to AI and get back commit organization plan."""
        try:
            # Ensure config is loaded
            if self.config is None:
                self.config = self.config_manager.load_config()
            
            ai_provider = UnifiedAIProvider(self.config)
            
            # Build prompt with individual hunks
            prompt = self._build_hunk_prompt(hunks, self.config.hunk.show_hunk_context)
            
            response = ai_provider.generate(prompt)
            
            # With structured output, response should always be valid JSON
            return json.loads(response)
            
        except json.JSONDecodeError as e:
            self.console.print(f"[red]AI returned invalid JSON: {e}[/red]")
            return None
        except Exception as e:
            self.console.print(f"[red]AI analysis failed: {e}[/red]")
            return None
    
    def _build_hunk_prompt(self, hunks: List[Hunk], show_context: bool = True) -> str:
        """Build AI prompt with individual hunks."""
        prompt_parts = [
            "Analyze these code changes and organize them into logical commits by hunk.",
            "",
            "For each commit, provide:",
            "1. A conventional commit message (type: description)", 
            "2. The specific hunk IDs that should be included",
            "3. A brief rationale for why these hunks belong together",
            "",
            "Return your response in this structure:",
            "{",
            '  "commits": [',
            "    {",
            '      "message": "feat: add user authentication system",',
            '      "hunk_ids": ["auth.py:45-89", "models.py:23-45"],',
            '      "rationale": "Groups authentication functionality together"',
            "    }",
            "  ]",
            "}",
            "",
            "Group related hunks together based on functionality, not just file location.",
            "Each hunk represents a specific change block that can be applied independently.",
            "",
            "HUNKS TO ANALYZE:",
            ""
        ]
        
        for hunk in hunks:
            hunk_parts = [
                f"Hunk ID: {hunk.id}",
                f"File: {hunk.file_path}"
            ]
            
            if show_context and hunk.context:
                hunk_parts.extend([
                    "Context:",
                    hunk.context,
                    ""
                ])
            
            hunk_parts.extend([
                "Changes:",
                hunk.content,
                "",
                "---",
                ""
            ])
            
            prompt_parts.extend(hunk_parts)
        
        return "\n".join(prompt_parts)
    
    def display_commit_plan_hunks(self, commit_plan: List[Dict[str, Any]], hunk_map: Dict[str, Hunk]):
        """Display the proposed commit plan with hunks."""
        self.console.print("\n[bold]Proposed Commit Structure:[/bold]")
        
        for i, commit in enumerate(commit_plan, 1):
            panel_content = []
            panel_content.append(f"[bold]Message:[/bold] {commit['message']}")
            
            # Display hunks grouped by file for readability
            if commit.get('hunk_ids'):
                hunk_ids = commit['hunk_ids']
                files_hunks = {}
                
                # Group hunk IDs by file
                for hunk_id in hunk_ids:
                    if hunk_id in hunk_map:
                        hunk = hunk_map[hunk_id]
                        if hunk.file_path not in files_hunks:
                            files_hunks[hunk.file_path] = []
                        files_hunks[hunk.file_path].append(hunk)
                
                panel_content.append("[bold]Hunks:[/bold]")
                for file_path, file_hunks in files_hunks.items():
                    panel_content.append(f"  📁 {file_path}")
                    for hunk in file_hunks:
                        line_range = f"lines {hunk.start_line}-{hunk.end_line}"
                        panel_content.append(f"    • {hunk.id} ({line_range})")
            
            panel_content.append(f"[bold]Rationale:[/bold] {commit['rationale']}")
            
            self.console.print(Panel(
                "\n".join(panel_content),
                title=f"Commit #{i}",
                border_style="blue"
            ))
    
    def get_user_confirmation(self) -> bool:
        """Get user confirmation to proceed."""
        self.console.print("\n[bold]Apply this commit structure?[/bold]")
        response = input("Continue? (y/N): ")
        return response.lower().strip() == 'y'
    
    def apply_commit_plan_hunks(self, commit_plan: List[Dict[str, Any]], hunk_map: Dict[str, Hunk], base_branch: str):
        """Apply the commit plan using hunk-based staging."""
        try:
            with Progress(
                SpinnerColumn(),
                TextColumn("[progress.description]{task.description}"),
                console=self.console
            ) as progress:
                # 1. Create backup branch
                task = progress.add_task("Creating backup...", total=None)
                current_branch = subprocess.run(
                    ['git', 'rev-parse', '--abbrev-ref', 'HEAD'],
                    capture_output=True, text=True, check=True
                ).stdout.strip()
                
                backup_branch = f"{current_branch}-backup-{int(__import__('time').time())}"
                subprocess.run(['git', 'branch', backup_branch], check=True)
                self.console.print(f"[green]Created backup branch: {backup_branch}[/green]")
                
                # 2. Reset to base branch
                progress.update(task, description="Resetting to base branch...")
                reset_to_base_branch(base_branch)
                
                # 3. Create new commits based on the plan
                progress.update(task, description="Creating new commits...")
                
                if commit_plan:
                    commits_created = 0
                    applied_hunk_ids = set()
                    
                    for i, commit in enumerate(commit_plan):
                        progress.update(task, description=f"Creating commit {i+1}/{len(commit_plan)}: {commit['message'][:50]}...")
                        
                        # Get hunk IDs for this commit
                        hunk_ids = commit.get('hunk_ids', [])
                        if hunk_ids:
                            # Verify hunks exist and haven't been applied yet
                            valid_hunk_ids = []
                            for hunk_id in hunk_ids:
                                if hunk_id in hunk_map and hunk_id not in applied_hunk_ids:
                                    valid_hunk_ids.append(hunk_id)
                                elif hunk_id in applied_hunk_ids:
                                    self.console.print(f"[yellow]Hunk {hunk_id} already applied, skipping[/yellow]")
                                else:
                                    self.console.print(f"[yellow]Hunk {hunk_id} not found, skipping[/yellow]")
                            
                            if valid_hunk_ids:
                                try:
                                    # Apply the hunks for this commit
                                    apply_hunks_by_ids(valid_hunk_ids, hunk_map)
                                    
                                    # Create the commit
                                    subprocess.run([
                                        'git', 'commit', '-m', commit['message']
                                    ], check=True)
                                    commits_created += 1
                                    applied_hunk_ids.update(valid_hunk_ids)
                                    
                                except HunkApplicatorError as e:
                                    self.console.print(f"[yellow]Failed to apply hunks for commit '{commit['message']}': {e}[/yellow]")
                                    # Continue with next commit
                                    continue
                            else:
                                self.console.print(f"[yellow]Skipping commit '{commit['message']}' - no valid hunks to apply[/yellow]")
                        else:
                            self.console.print(f"[yellow]Skipping commit '{commit['message']}' - no hunk IDs specified[/yellow]")
                    
                    # Check if there are any hunks that weren't applied
                    unapplied_hunks = set(hunk_map.keys()) - applied_hunk_ids
                    if unapplied_hunks:
                        progress.update(task, description="Creating final commit for remaining hunks...")
                        try:
                            apply_hunks_by_ids(list(unapplied_hunks), hunk_map)
                            subprocess.run([
                                'git', 'commit', '-m', 'chore: remaining uncommitted changes'
                            ], check=True)
                            commits_created += 1
                        except HunkApplicatorError as e:
                            self.console.print(f"[yellow]Failed to apply remaining hunks: {e}[/yellow]")
                    
                    self.console.print(f"[green]Successfully created {commits_created} new commit(s)[/green]")
                    self.console.print(f"[blue]Backup available at: {backup_branch}[/blue]")
                
        except subprocess.CalledProcessError as e:
            self.console.print(f"[red]Failed to apply commit plan: {e}[/red]")
            sys.exit(1)
        except Exception as e:
            self.console.print(f"[red]Error applying commit plan: {e}[/red]")
            sys.exit(1)


def main():
    """Entry point for the git-smart-squash command."""
    cli = GitSmartSquashCLI()
    cli.main()


if __name__ == '__main__':
    main()